// helptext.c

// Online help for PM Preference

// Todd Fincannon
// BIX: tfincannon
// CIS: 70526,1742
// Usenet: tlf@well.sf.ca.us  {ucbvax,apple,hplabs,pacbell}!well!tlf

// Revision log:
// 1.00 |  3 Apr 90 | initial release

// Compilation:
// Using Microsoft C 5.1 or later,
// with the Microsoft PM Toolkit 1.1 or later,
// cl -c -AL -G2sw -W3 -Ot pmpref.c
// cl -c -AL -G2sw -W3 -Ot helptext.c
// rc -r pmpref
// link pmpref helptext, /align:16, NUL, os2, pmpref
// rc pmpref.res


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                        P R E L I M I N A R I E S                         //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

#define INCL_DOSMISC
#define INCL_DOSRESOURCES
#define INCL_GPILCIDS
#define INCL_GPILOGCOLORTABLE
#define INCL_WINFRAMEMGR
#define INCL_WINHOOKS
#define INCL_WININPUT
#define INCL_WINSCROLLBARS
#define INCL_WINSYS
#include <os2.h>

#include <stdlib.h>
#include <ctype.h>
#include "pmpref.h"

VOID HelpInit (HAB hab, HMQ hmq);
VOID HelpTerminate (HAB hab, HMQ hmq);

MRESULT CALLBACK HelpWndProc (HWND, USHORT, MPARAM, MPARAM);

static VOID ShowHelp (USHORT idHelpText);
static VOID SaveHelpWindowPos (VOID);
static VOID SetHelpColors (PLONG alHelpColors);
static VOID BreakTextLines (PCHAR pchHelpText, ULONG ulHelpSize,
                            SHORT sMaxWidth, LONG alWidth [],
                            TEXTLINE atextline [], SHORT *psNumLines);

BOOL fHelpIsShown = FALSE;       // True when the help window is displayed
BOOL fHelpIsMaxed = FALSE;       // True then the help window is maximized

SHORT sHelpBLCx, sHelpBLCy;      // Position of the help window
SHORT cxHelpFrame, cyHelpFrame;  // Size of the help window

HWND hwndHelp;                   // Help client window handle
HWND hwndHelpFrame;              // Help frame window handle

CHAR szHelpClass [] = "HelpText";


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                            H E L P   H O O K                             //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

BOOL CALLBACK HelpHook (HAB hab, USHORT usMode, USHORT idTopic,
                        USHORT idSubTopic, PRECTL prcPosition) {

   // The help hook is the entry point for all help requests.
   // It is called when the user presses F1 or selects
   // "PM Preference Help" from the Help menu.

   // The text resource identifier (IDT_) for each help topic must
   // be unique and must be the same as the help hook subtopic ID.

   // If the user presses F1 without selecting a menu item or
   // selects help from the Help menu, display general help.

   if (usMode == HLPM_WINDOW && idSubTopic == FID_CLIENT) {
      ShowHelp (IDT_HELP);
      return TRUE;
   }

   // The user pressed F1 while a menu item was selected.

   if (usMode == HLPM_MENU) {

      switch (idSubTopic) {

         case IDT_RESTORE:
         case IDT_UNDO:
         case IDT_SAVE:
         case IDT_SAVEAS:
         case IDT_DELETE:
         case IDT_HELP:
         case IDT_ABOUT:

            // Help for specific menu items.
            ShowHelp (idSubTopic);
            break;

         case SC_CLOSE:

            // Help for the Exit menu item.
            ShowHelp (IDT_EXIT);
            break;

         case 0xFFFF:

            // Help for whole menus.
            switch (idTopic) {

               case IDT_PREFMENU:
               case IDT_HELPMENU:
                  ShowHelp (idTopic);
                  break;
            }
      }
   }

   return TRUE;
}


VOID ShowHelp (USHORT idHelpText) {

   // Pop up a help text window for the indicated item,
   // or replace the text in an already-displayed window.

   static ULONG flHelpFCF = FCF_TITLEBAR | FCF_SYSMENU | FCF_MAXBUTTON |
                            FCF_SIZEBORDER | FCF_VERTSCROLL;
   USHORT fsWinPos;

   if (fHelpIsShown) {

      // If the help window is already displayed, just change the text.
      WinSendMsg (hwndHelpFrame, WM_CHANGEHELPTEXT, MPFROMSHORT (idHelpText), 0);

      // Bring the help window to the top.
      WinSetActiveWindow (HWND_DESKTOP, hwndHelpFrame);
      return;
   }

   // Create a standard window but do not display it.
   fHelpIsShown = TRUE;
   hwndHelpFrame = WinCreateStdWindow (HWND_DESKTOP, 0L,
                                      &flHelpFCF, szHelpClass,
                                      "PM Preference Help",
                                      0L, 0, 0, &hwndHelp);

   // Load the help text resource.
   WinSendMsg (hwndHelpFrame, WM_CHANGEHELPTEXT, MPFROMSHORT (idHelpText), 0);

   // If the help window was shown previously, use the size
   // that was set then.  Otherwise, calculate a default size.

   if (cxHelpFrame == 0) {

      SHORT sWidth, sHeight;
      SHORT cxBorder = (SHORT) WinQuerySysValue (HWND_DESKTOP, SV_CXSIZEBORDER);
      SHORT cyBorder = (SHORT) WinQuerySysValue (HWND_DESKTOP, SV_CYSIZEBORDER);
      SHORT cyTitleBar = (SHORT) WinQuerySysValue (HWND_DESKTOP, SV_CYTITLEBAR);
      SHORT cyChar;
      RECTL rclDesktop;
      FONTMETRICS fm;
      HPS hps;

      // Get the boundary rectangle of the desktop.
      WinQueryWindowRect (HWND_DESKTOP, &rclDesktop);
      sHeight = (SHORT) (rclDesktop.yTop - rclDesktop.yBottom);
      sWidth = (SHORT) (rclDesktop.xRight - rclDesktop.xLeft);

      // Get the font metrics for the default font.
      hps = WinGetPS (HWND_DESKTOP);
      GpiQueryFontMetrics (hps, (LONG) sizeof fm, &fm);
      cyChar = (SHORT) fm.lMaxBaselineExt;
      WinReleasePS (hps);

      // Set the position and size of the help window.
      sHelpBLCx = sWidth / 2 - cxBorder;
      sHelpBLCy = sHeight / 4;
      cxHelpFrame = sWidth / 2;
      cyHelpFrame = 5 * sHeight / 8;

      // Adjust the frame size so that the client window vertical dimension
      // is a multiple of the font size (no partial lines displayed).
      cyHelpFrame -= (cyHelpFrame - cyTitleBar - cyBorder * 2) % cyChar;
   }

   // Set window positioning flags.
   fsWinPos = SWP_ACTIVATE | SWP_SHOW | SWP_SIZE | SWP_MOVE | SWP_ZORDER;

   // Maximize the window if it was maximized the last time it was shown.
   if (fHelpIsMaxed)
      fsWinPos |= SWP_MAXIMIZE;

   // Set the size and position of the window and show it.
   WinSetWindowPos (hwndHelpFrame, HWND_TOP, sHelpBLCx, sHelpBLCy,
                    cxHelpFrame, cyHelpFrame, fsWinPos);
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                H E L P   W I N D O W   P R O C E D U R E                 //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

MRESULT CALLBACK HelpWndProc (HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2) {

   static HWND hwndVscroll;
   static SHORT sVscrollMax, sVscrollPos;
   static SHORT cyChar, cyDesc;
   static SHORT cxClient, cyClient;
   static SHORT sMaxWidth;
   static PCHAR pchHelpText = NULL;
   static SEL selHelpText;
   static SHORT sNumLines = 0;
   static ULONG ulHelpSize = 0;
   static TEXTLINE atextline [NUM_TEXTLINES];
   static LONG alWidth [NUM_PRINT_CHARS];
   static LONG alHelpColors [3];
   static BOOL fTrackSlider = FALSE;

   FONTMETRICS fm;
   HPS hps;
   POINTL ptl;
   SHORT sLine, sPaintBeg, sPaintEnd, sVscrollInc;
   RECTL rclInvalid;
   USHORT idHelpText;
   USHORT usOSVersion;

   switch (msg) {

   case WM_CREATE:
      // Get the font metrics for the client window presentation space.
      hps = WinGetPS (hwnd);
      GpiQueryFontMetrics (hps, (LONG) sizeof fm, &fm);
      cyChar = (SHORT) fm.lMaxBaselineExt;
      cyDesc = (SHORT) fm.lMaxDescender;
      GpiQueryWidthTable (hps, 0x20L, (LONG) NUM_PRINT_CHARS, alWidth);
      WinReleasePS (hps);

      // Set the help window colors.
      SetHelpColors (alHelpColors);

      // Save the vertical scroll bar window handle for later use.
      hwndVscroll = WinWindowFromID (WinQueryWindow (hwnd, QW_PARENT, FALSE),
                                     FID_VERTSCROLL);

      // Get the operation system version number.
      DosGetVersion (&usOSVersion);
      // Track the scroll bar slider only if we're running OS/2 1.2 or above.
      fTrackSlider = HIBYTE (usOSVersion) >= 10 && LOBYTE (usOSVersion) > 10;

      return MRFROMSHORT (FALSE);

   case WM_DESTROY:
      if (pchHelpText != NULL) {
         DosFreeSeg (selHelpText);
         pchHelpText = NULL;
      }
      return MRFROMSHORT (FALSE);

   case WM_CLOSE:
      fHelpIsShown = FALSE;
      WinDestroyWindow (hwndHelpFrame);
      return MRFROMSHORT (FALSE);

   case WM_CHANGEHELPTEXT:
      //
      // The WM_CHANGEHELPTEXT custom message indicates that the help
      // text resource selection has changed.  It is sent when the
      // help window is first created and whenever the text changes
      // afterward.
      //
      // Message parameters:
      // idHelpText = SHORT1FROMMP (mp1);  // text resource id
      //

      idHelpText = SHORT1FROMMP (mp1);

      // If a help text resource is already loaded, free it.
      if (pchHelpText != NULL) {
         DosFreeSeg (selHelpText);
         pchHelpText = NULL;
      }

      // Load the help text resource and display it in the help window.
      if (!DosGetResource (0, IDT_TEXT, idHelpText, &selHelpText)) {

         // Get the new segment size and a pointer to it.
         DosSizeSeg (selHelpText, &ulHelpSize);
         pchHelpText = MAKEP (selHelpText, 0);

         // Fill in the atextline array with line breaks based on the
         // window size.  The number of lines is returned in sNumLines.
         BreakTextLines (pchHelpText, ulHelpSize, sMaxWidth,
                         alWidth, atextline, &sNumLines);

         // Show the text.
         WinInvalidateRect (hwnd, NULL, FALSE);

         // Set the range of the scroll bar and the position of the slider.
         sVscrollMax = max (0, sNumLines - cyClient / cyChar);
         sVscrollPos = 0;
         WinSendMsg (hwndVscroll, SBM_SETSCROLLBAR,
                     MPFROM2SHORT (sVscrollPos, 0),
                     MPFROM2SHORT (0, sVscrollMax));

         // Turn off the slider if the window can display all the text.
         WinEnableWindow (hwndVscroll, sVscrollMax ? TRUE : FALSE);
      }

      else {
         // The text resource was not found, so close the help window.
         WinSendMsg (hwndHelp, WM_CLOSE, 0, 0);
      }

      return MRFROMSHORT (FALSE);

   case WM_MOVE:
      // Save the size and position of the help frame window.
      SaveHelpWindowPos ();
      return MRFROMSHORT (FALSE);

   case WM_MINMAXFRAME:
      if (((PSWP) mp1)->fs & SWP_MAXIMIZE)
         fHelpIsMaxed = TRUE;
      else
         fHelpIsMaxed = FALSE;
      return MRFROMSHORT (FALSE);

   case WM_SYSCOLORCHANGE:
      SetHelpColors (alHelpColors);
      return NULL;

   case WM_SIZE:
      // Get the size of the client window.
      cxClient = SHORT1FROMMP (mp2);
      cyClient = SHORT2FROMMP (mp2);

      // Save the size and position of the help frame window.
      SaveHelpWindowPos ();

      // The maximum line width is the client window size
      // minus border areas on the left and right.
      sMaxWidth = cxClient - 2 * (SHORT) alWidth [WNDX('y')];

      // Fill in the atextline array with line breaks based on the
      // new window size.  The number of lines is returned in sNumLines.
      BreakTextLines (pchHelpText, ulHelpSize, sMaxWidth,
                      alWidth, atextline, &sNumLines);

      // Set the range of the scroll bar and the position of the slider.
      sVscrollMax = max (0, sNumLines - cyClient / cyChar);
      sVscrollPos = min (sVscrollPos, sVscrollMax);
      WinSendMsg (hwndVscroll, SBM_SETSCROLLBAR,
                  MPFROM2SHORT (sVscrollPos, 0),
                  MPFROM2SHORT (0, sVscrollMax));

      // Turn off the slider if the window can display all the text.
      WinEnableWindow (hwndVscroll, sVscrollMax ? TRUE : FALSE);

      return MRFROMSHORT (FALSE);

   case WM_VSCROLL:

      // Handle vertical scroll bar actions.

      switch (SHORT2FROMMP (mp2)) {

         case SB_LINEUP:
            sVscrollInc = -1;
            break;

         case SB_LINEDOWN:
            sVscrollInc = 1;
            break;

         case SB_PAGEUP:
            sVscrollInc = min (-1, -cyClient / cyChar);
            break;

         case SB_PAGEDOWN:
            sVscrollInc = max (1, cyClient / cyChar);
            break;

         case SB_SLIDERTRACK:
            sVscrollInc = SHORT1FROMMP (mp2) - sVscrollPos;
            break;

         case SB_SLIDERPOSITION:
            // If we weren't tracking the slider because of the bug in 1.1,
            // set the slider position now that dragging is finished.

            if (!fTrackSlider)
               WinSendMsg (hwndVscroll, SBM_SETPOS,
                           MPFROMSHORT (sVscrollPos), NULL);
            return MRFROMSHORT (FALSE);

         default:
            sVscrollInc = 0;
            break;
      }

      // Decide how much to scroll the text.
      sVscrollInc = max (-sVscrollPos,
                    min (sVscrollInc, sVscrollMax - sVscrollPos));

      // Scroll the text and move the scroll bar slider.
      if (sVscrollInc != 0) {
         sVscrollPos += sVscrollInc;
         WinScrollWindow (hwnd, 0, cyChar * sVscrollInc,
                          NULL, NULL, NULL, NULL, SW_INVALIDATERGN);

         // OS/2 1.1 does not update the scroll bar properly when the
         // slider is dragged using the mouse.  If the system is
         // running 1.1 and the slider is being dragged, wait until
         // the user finishes dragging the slider to set its position.

         if (fTrackSlider || SHORT2FROMMP (mp2) != SB_SLIDERTRACK)
            WinSendMsg (hwndVscroll, SBM_SETPOS, MPFROMSHORT (sVscrollPos),
                        NULL);

         WinUpdateWindow (hwnd);
      }
      return MRFROMSHORT (FALSE);

   case WM_CHAR:

      switch (CHARMSG(&msg)->vkey) {

         case VK_UP:
         case VK_DOWN:
         case VK_PAGEUP:
         case VK_PAGEDOWN:
            // Translate cursor movement keystrokes into scroll bar actions.
            return WinSendMsg (hwndVscroll, msg, mp1, mp2);

         case VK_ESC:
         case VK_F3:
            // Close the help window if the user presses Escape or F3.
            WinSendMsg (hwndHelp, WM_CLOSE, 0, 0);
            return MRFROMSHORT (TRUE);
      }
      break;

   case WM_PAINT:
      // Set text colors for the help window.
      hps = WinBeginPaint (hwnd, NULL, &rclInvalid);
      GpiCreateLogColorTable (hps, 0L, LCOLF_CONSECRGB, CLR_BACKGROUND, 3L,
                              alHelpColors);
      GpiSetColor (hps, CLR_HELPTEXT);

      // Erase the invalid area.
      GpiErase (hps);

      // Find the start and end lines encompassed by the invalid area.
      sPaintBeg = max (0, sVscrollPos +
                       (cyClient - (SHORT) rclInvalid.yTop) / cyChar);
      sPaintEnd = min (sNumLines, sVscrollPos +
                       (cyClient - (SHORT) rclInvalid.yBottom) / cyChar + 1);

      // Display each line using stored text pointers and line lengths.
      for (sLine = sPaintBeg; sLine < sPaintEnd; sLine++) {

         ptl.x = alWidth [WNDX ('y')];
         ptl.y = cyClient - cyChar * (sLine + 1 - sVscrollPos) + cyDesc;

         GpiCharStringAt (hps, &ptl, (LONG) atextline [sLine].sLength,
                          atextline [sLine].szText);
      }

      WinEndPaint (hps);
      return MRFROMSHORT (FALSE);
   }
   return WinDefWindowProc (hwnd, msg, mp1, mp2);
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//               H E L P   U T I L I T Y   F U N C T I O N S                //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

VOID HelpInit (HAB hab, HMQ hmq) {

   // This function must be called by the main program
   // to install online help.

   // Register the help window class.
   WinRegisterClass (hab, szHelpClass, HelpWndProc,
                     CS_SIZEREDRAW | CS_MOVENOTIFY, 0);

   // Install the help hook for menu items.
   WinSetHook (hab, hmq, HK_HELP, (PFN) HelpHook, 0);
}


VOID HelpTerminate (HAB hab, HMQ hmq) {

   // This function must be called by the main program
   // when it terminates to release the help hook.

   WinReleaseHook (hab, hmq, HK_HELP, (PFN) HelpHook, 0);
}


VOID SaveHelpWindowPos (VOID) {

   // Save the size and position of the help frame window.
   // Do not overwrite the current size and position if
   // the window is being maximized.

   SWP swpHelp;

   WinQueryWindowPos (hwndHelpFrame, &swpHelp);
   if (!(swpHelp.fs & SWP_MAXIMIZE)) {
      sHelpBLCx = swpHelp.x;
      sHelpBLCy = swpHelp.y;
      cxHelpFrame = swpHelp.cx;
      cyHelpFrame = swpHelp.cy;
   }
}


VOID SetHelpColors (PLONG alHelpColors) {

   // Set the help window colors from the current system colors.

   alHelpColors [0] = WinQuerySysColor (HWND_DESKTOP, SYSCLR_HELPBACKGROUND, 0L);
   alHelpColors [1] = WinQuerySysColor (HWND_DESKTOP, SYSCLR_HELPTEXT, 0L);
   alHelpColors [2] = WinQuerySysColor (HWND_DESKTOP, SYSCLR_HELPHILITE, 0L);
}


VOID BreakTextLines (PCHAR pchHelpText, ULONG ulHelpSize, SHORT sMaxWidth,
                     LONG alWidth [], TEXTLINE atextline [],
                     SHORT *psNumLines) {

   // This routine takes the text indicated by pchHelpText, with a
   // length of ulHelpSize, and breaks it into lines based on the
   // width of the display window, given by sMaxWidth.  It fills in
   // a TEXTLINE structure with a pointer to the start of each line
   // and the length of the line in characters for each display line.
   // Line lengths are computed based on the alWidth character table.
   // The number of text lines is returned in psNumLines.

   PCHAR pchLine, pchPrevWord, pchCurrent;
   SHORT sLineWidth, sPrevWidth, sLine;

   if (pchHelpText != NULL && ulHelpSize > 0L) {

      // Start out with a blank line at the beginning of the text.
      atextline [0].szText = pchHelpText;
      pchLine = pchHelpText;
      pchPrevWord = pchHelpText;
      pchCurrent = pchHelpText;
      sLineWidth = 0;
      sPrevWidth = 0;
      sLine = 0;

      // The following loop fetches one character per iteration,
      // filling in a TEXTLINE structure at each line break.

      // The end of a line comes when:
      // 1. the loop terminates, or
      // 2. we encounter a carriage return character, or
      // 3. the latest word puts us over the window width.

      // The loop terminates when:
      // 1. we run out of characters, or
      // 2. we encounter an end-of-file character, or
      // 3. we run out of lines to fill.

      do {
         // Add this character's width to the total for this line.
         sLineWidth += (SHORT) alWidth [WNDX (*pchCurrent)];

         // Get to the end of the current word.
         if (isspace (*pchCurrent++)) {

            // See if we have run out of room yet.
            if (sLineWidth > sMaxWidth) {

               // Break the line at the previous word.
               // Set the length of the current line.
               atextline [sLine].sLength = pchPrevWord - pchLine - 1;

               // Set the start of the new line.
               sLine++;
               pchLine = pchPrevWord;
               atextline [sLine].szText = pchLine;
               sLineWidth -= sPrevWidth;
            }

            // If the current character is a carriage return,
            // break the line here.
            if (pchCurrent [-1] == '\r') {
               atextline [sLine].sLength = pchCurrent - pchLine - 1;
               sLine++;
               if (*pchCurrent == '\n')
                  pchCurrent++;
               pchLine = pchCurrent;
               atextline [sLine].szText = pchLine;
               sLineWidth = 0;
            }

            // Set the current word as the previous word.
            pchPrevWord = pchCurrent;
            sPrevWidth = sLineWidth;
         }
      }
      // Continue while there are more characters and room for lines.
      while (pchCurrent - pchHelpText < (SHORT) ulHelpSize &&
             *pchCurrent != EOF_CHAR &&
             sLine < NUM_TEXTLINES - 1);

      // Check if the last word went out of bounds.
      if (sLineWidth > sMaxWidth) {

         atextline [sLine].sLength = pchPrevWord - pchLine - 1;

         // Set the start of the new line.
         if (sLine < NUM_TEXTLINES - 1) {
            sLine++;
            pchLine = pchPrevWord;
            atextline [sLine].szText = pchLine;
         }
      }

      // Set the total line count and length of the previous line.
      atextline [sLine].sLength = pchCurrent - pchLine;
      *psNumLines = min (sLine + 1, NUM_TEXTLINES);
   }

   else {
      // There was no text, so set the number of lines to zero.
      *psNumLines = 0;
   }
}
