// pmpref.c

// Save and restore Presentation Manager preference settings.

// Todd Fincannon
// BIX: tfincannon
// CIS: 70526,1742
// Usenet: tlf@well.sf.ca.us  {ucbvax,apple,hplabs,pacbell}!well!tlf

// Revision log:
// 1.00 |  3 Apr 90 | initial release

// Compilation:
// Using Microsoft C 5.1 or later,
// with the Microsoft PM Toolkit 1.1 or later,
// cl -c -AL -G2sw -W3 -Ot pmpref.c
// cl -c -AL -G2sw -W3 -Ot helptext.c
// rc -r pmpref
// link pmpref helptext, /align:16, NUL, os2, pmpref
// rc pmpref.res


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                        P R E L I M I N A R I E S                         //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

#define INCL_DOSFILEMGR
#define INCL_DOSMEMMGR
#define INCL_DOSMISC
#define INCL_DOSPROCESS
#define INCL_GPIBITMAPS
#define INCL_GPICONTROL
#define INCL_GPILOGCOLORTABLE
#define INCL_GPIPRIMITIVES
#define INCL_WINBUTTONS
#define INCL_WINDIALOGS
#define INCL_WINENTRYFIELDS
#define INCL_WINERRORS
#define INCL_WINFRAMEMGR
#define INCL_WININPUT
#define INCL_WINLISTBOXES
#define INCL_WINMENUS
#define INCL_WINMESSAGEMGR
#define INCL_WINPOINTERS
#define INCL_WINSCROLLBARS
#define INCL_WINSHELLDATA
#define INCL_WINSWITCHLIST
#define INCL_WINSYS
#define INCL_WINWINDOWMGR
#include <os2.h>

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "pmpref.h"

// PM Preference main function

VOID cdecl main (int argc, char *argv[]);

// Window and dialog procedures

MRESULT CALLBACK PrefWndProc (HWND, USHORT, MPARAM, MPARAM);
MRESULT CALLBACK RestoreDlgProc (HWND, USHORT, MPARAM, MPARAM);
MRESULT CALLBACK SaveAsDlgProc (HWND, USHORT, MPARAM, MPARAM);
MRESULT CALLBACK DeleteDlgProc (HWND, USHORT, MPARAM, MPARAM);
MRESULT CALLBACK AboutDlgProc  (HWND, USHORT, MPARAM, MPARAM);

// Online help initialization and termination, defined in file helptext.c

VOID HelpInit (HAB hab, HMQ hmq);
VOID HelpTerminate (HAB hab, HMQ hmq);

// PM Preference low-level functions

static VOID GetPrefValues (SHORT ipref);
static VOID SetPrefValues (SHORT ipref);
static VOID ReadPrefFile (VOID);
static VOID WritePrefFile (VOID);
static VOID WriteIni (SHORT ipref);

// PM Preference utility functions

static VOID ChangeCurrentPref (VOID);
static VOID SaveNewPref (HWND hwnd);
static VOID SetValue (SHORT isSysValue, LONG lValue);
static VOID FillPrefListBox (HWND hwnd, USHORT idListBox);
static SHORT FindPrefName (CHAR *pchPrefName);
static SHORT FindPrefKey (CHAR *pchKey);
static BOOL PrefAlloc (VOID);

// General utility functions

static VOID AbortPgm (CHAR *szErrorMessage);
static VOID DisplayMsg (CHAR *szErrorMessage);
static VOID DisplayError (VOID);
static CHAR *StrTrimLeft (CHAR *pch);
static CHAR *StrTrimRight (CHAR *pch);
static VOID ShowHourglassPointer (VOID);
static VOID ShowNormalPointer (VOID);
static VOID CenterDialogBox (HWND hwnd);

// The "pref list" (pointed to by pprefList) is an array of preference
// structures.  Each one contains the system colors and system values
// controlled by the Preferences menu in the Control Panel program.
// The first preference (index number 0) holds the settings in effect
// when the program started, so users can always get back where they
// started.  Thus the preferences read from the preference file start
// at index number 1.

PPREF pprefList = NULL;

// cpref counts the number of preferences in pprefList.  It starts
// at one because the list holds the original preference at minimum.

SHORT cpref = 1;

// sMaxPref is the number of preferences for which we have allocated memory.

SHORT sMaxPref = 0;


SHORT iprefPrevious = 0;      // The index of the previous preference
SHORT iprefCurrent = 0;       // The index of the current preference

BOOL fEmptyPrefList = TRUE;   // True when the pref list is empty
BOOL fPermanentPref = FALSE;  // True when a restored pref should be permanent
BOOL fUnsavedChanges = FALSE; // True when delete is done without save
BOOL fExtraValues = FALSE;    // True when OS/2 1.2 or above is running

HWND hwndPref;                // Preference client window handle
HWND hwndPrefFrame;           // Preference frame window handle

HAB hab;                      // Anchor block handle (whatever that is)

CHAR szPrefClass [] = "PMpref";        // Preference client window class name
CHAR szPrefTitle [] = "PM Preference"; // Initial title bar text
CHAR szPrefFileName [256];             // Fully qualified pathname of pref file
CHAR szCurrentPref [MAX_NAME_SIZE];    // The name of the current preference
CHAR szNewPref [MAX_NAME_SIZE];        // The name of a new preference

CHAR szTitlebar [MAX_NAME_SIZE + sizeof szPrefTitle + 3];  // Titlebar text


// Default preference values for OS/2 1.2

PREF prefDefault = {
   0xffffff,   // ButtonLight
   0xcccccc,   // ButtonMiddle
   0x808080,   // ButtonDark
   0x000000,   // ButtonDefault
   0xcccccc,   // TitleBottom
   0x808080,   // Shadow
   0x000000,   // IconText
   0xffffff,   // DialogBackground
   0xffffff,   // HiliteForeground
   0x000000,   // HiliteBackground
   0x000000,   // InactiveTitleTextBgnd
   0x000000,   // ActiveTitleTextBgnd
   0x000000,   // InactiveTitleText
   0xffffff,   // ActiveTitleText
   0x000000,   // OutputText
   0x000080,   // WindowStaticText
   0xe0e0e0,   // Scrollbar
   0xcccccc,   // Background
   0x004080,   // ActiveTitle
   0xcccccc,   // InactiveTitle
   0xffffff,   // Menu
   0xffffff,   // Window
   0x808080,   // WindowFrame
   0x000000,   // MenuText
   0x000000,   // WindowText
   0xffffff,   // TitleText
   0xffff00,   // ActiveBorder
   0xcccccc,   // InactiveBorder
   0xffffff,   // AppWorkspace
   0xffffff,   // HelpBackground
   0x00007f,   // HelpText
   0x007f7f,   // HelpHilite
         4L,   // BorderWidth
         1L,   // Beep
       700L,   // CursorBlinkRate
       500L,   // DoubleClickSpeed
         0L,   // SwapMouseButtons
       116L    // TaskListMouseAccess
};


// The aprefindex array is used to look up the os2.ini name string and
// identifying constant for each system color and system value.  It also
// gives an index into one of the two arrays in the PREF structure that
// holds color values (PI_COLOR) or system values (PI_VALUE).  Two colors
// are not stored in os2.ini by OS/2, so I added two new setting strings
// (InactiveTitleTextBgnd and ActiveTitleTextBgnd).

PREFINDEX aprefindex [] = {
   "ButtonLight",          PI_COLOR,    0,   SYSCLR_BUTTONLIGHT,
   "ButtonMiddle",         PI_COLOR,    1,   SYSCLR_BUTTONMIDDLE,
   "ButtonDark",           PI_COLOR,    2,   SYSCLR_BUTTONDARK,
   "ButtonDefault",        PI_COLOR,    3,   SYSCLR_BUTTONDEFAULT,
   "TitleBottom",          PI_COLOR,    4,   SYSCLR_TITLEBOTTOM,
   "Shadow",               PI_COLOR,    5,   SYSCLR_SHADOW,
   "IconText",             PI_COLOR,    6,   SYSCLR_ICONTEXT,
   "DialogBackground",     PI_COLOR,    7,   SYSCLR_DIALOGBACKGROUND,
   "HiliteForeground",     PI_COLOR,    8,   SYSCLR_HILITEFOREGROUND,
   "HiliteBackground",     PI_COLOR,    9,   SYSCLR_HILITEBACKGROUND,
   "InactiveTitleTextBgnd",PI_COLOR,   10,   SYSCLR_INACTIVETITLETEXTBGND,
   "ActiveTitleTextBgnd",  PI_COLOR,   11,   SYSCLR_ACTIVETITLETEXTBGND,
   "InactiveTitleText",    PI_COLOR,   12,   SYSCLR_INACTIVETITLETEXT,
   "ActiveTitleText",      PI_COLOR,   13,   SYSCLR_ACTIVETITLETEXT,
   "OutputText",           PI_COLOR,   14,   SYSCLR_OUTPUTTEXT,
   "WindowStaticText",     PI_COLOR,   15,   SYSCLR_WINDOWSTATICTEXT,
   "Scrollbar",            PI_COLOR,   16,   SYSCLR_SCROLLBAR,
   "Background",           PI_COLOR,   17,   SYSCLR_BACKGROUND,
   "ActiveTitle",          PI_COLOR,   18,   SYSCLR_ACTIVETITLE,
   "InactiveTitle",        PI_COLOR,   19,   SYSCLR_INACTIVETITLE,
   "Menu",                 PI_COLOR,   20,   SYSCLR_MENU,
   "Window",               PI_COLOR,   21,   SYSCLR_WINDOW,
   "WindowFrame",          PI_COLOR,   22,   SYSCLR_WINDOWFRAME,
   "MenuText",             PI_COLOR,   23,   SYSCLR_MENUTEXT,
   "WindowText",           PI_COLOR,   24,   SYSCLR_WINDOWTEXT,
   "TitleText",            PI_COLOR,   25,   SYSCLR_TITLETEXT,
   "ActiveBorder",         PI_COLOR,   26,   SYSCLR_ACTIVEBORDER,
   "InactiveBorder",       PI_COLOR,   27,   SYSCLR_INACTIVEBORDER,
   "AppWorkspace",         PI_COLOR,   28,   SYSCLR_APPWORKSPACE,
   "HelpBackground",       PI_COLOR,   29,   SYSCLR_HELPBACKGROUND,
   "HelpText",             PI_COLOR,   30,   SYSCLR_HELPTEXT,
   "HelpHilite",           PI_COLOR,   31,   SYSCLR_HELPHILITE,
   "BorderWidth",          PI_VALUE,    0,   SV_CXSIZEBORDER,
   "Beep",                 PI_VALUE,    1,   SV_ALARM,
   "CursorBlinkRate",      PI_VALUE,    2,   SV_CURSORRATE,
   "DoubleClickSpeed",     PI_VALUE,    3,   SV_DBLCLKTIME,
   "SwapMouseButtons",     PI_VALUE,    4,   SV_SWAPBUTTON,
   "TaskListMouseAccess",  PI_VALUE,    5,   SV_TASKLISTMOUSEACCESS
};

// Array sizes
#define NUM_SETTINGS    (sizeof aprefindex / sizeof (PREFINDEX))
#define NUM_COLORS      (sizeof prefDefault.alColor / sizeof (LONG))
#define NUM_VALUES      (sizeof prefDefault.alValue / sizeof (LONG))

// Index boundaries between 1.1 colors and values and 1.2 additions
#define COLOR_BOUNDARY  15
#define VALUE_BOUNDARY   5


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//           P M   P R E F E R E N C E   M A I N   F U N C T I O N          //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

VOID cdecl main (int argc, char *argv[]) {

   HMQ hmq;
   QMSG qmsg;
   SWCNTRL swcntrl;
   USHORT usOSVersion;

   ULONG flPrefFCF = FCF_TITLEBAR      | FCF_SYSMENU |
                     FCF_SIZEBORDER    | FCF_MINMAX  |
                     FCF_SHELLPOSITION | FCF_ICON    |
                     FCF_ACCELTABLE    | FCF_MENU;

   // Get the operation system version number.
   DosGetVersion (&usOSVersion);
   // Only use extra colors and values when we're running OS/2 1.2 or above.
   fExtraValues = HIBYTE (usOSVersion) >= 10 && LOBYTE (usOSVersion) > 10;

   // Start our message queue.
   hab = WinInitialize (0);
   hmq = WinCreateMsgQueue (hab, 0);

   ShowHourglassPointer ();

   // Register the preference client window class.
   WinRegisterClass (hab, szPrefClass, PrefWndProc, 0L, 0);

   // Install online help.
   HelpInit (hab, hmq);

   // Create the client window but do not show it yet.
   hwndPrefFrame = WinCreateStdWindow (HWND_DESKTOP, 0L, &flPrefFCF,
                                       szPrefClass, szPrefTitle, 0L, 0,
                                       ID_RESOURCE, &hwndPref);

   // Add our program to the switch list.

   swcntrl.hwnd             = hwndPrefFrame;
   swcntrl.hwndIcon         = NULL;
   swcntrl.hprog            = NULL;
   swcntrl.idProcess        = 0;
   swcntrl.idSession        = 0;
   swcntrl.uchVisibility    = 0;
   swcntrl.fbJump           = 0;
   swcntrl.fReserved        = 0;
   strcpy (swcntrl.szSwtitle, szPrefTitle);
   WinAddSwitchEntry (&swcntrl);

   // Allocate memory for preferences to be read from disk.
   if (!PrefAlloc ()) {
      AbortPgm ("Not enough memory");
   }
   else {
      // Store the original preference settings in the pref list, index 0.
      GetPrefValues (0);

      // Read all the saved preferences from the preference file into pprefList.
      ReadPrefFile ();

      // Clear the preference title since none is yet selected.
      *szCurrentPref = EOS;

      // Perform special initialization when the pref list is nonempty.
      if (cpref > 1) {
         fEmptyPrefList = FALSE;

         // Check for a preference on the command line.
         if (argc > 1) {
            SHORT sResult = FindPrefName (argv[1]);

            if (sResult >= 0) {
               // The indicated preference exists, so set it and exit.
               iprefCurrent = sResult;
               ChangeCurrentPref ();
               WinPostMsg (hwndPrefFrame, WM_QUIT, 0, 0);
            }
            else {
               // The preference was not found.  Display message & continue.
               CHAR achBuf [256];
               sprintf (achBuf, "\"%s\" is not in the list of preferences.",
                        argv [1]);
               DisplayMsg (achBuf);
            }
         }
      }
   }

   // Now that all the dirty work is done, show the window, except
   // when we get the preference from the command line or can't
   // allocate enough memory.

   if (iprefCurrent == 0 && sMaxPref > 0) {
      WinShowWindow(hwndPrefFrame, TRUE);
      WinSetFocus(HWND_DESKTOP, hwndPref);
   }

   ShowNormalPointer ();

   // Event loop.
   while (WinGetMsg (hab, &qmsg, NULL, 0, 0))
      WinDispatchMsg (hab, &qmsg);

   // Clean up everything and exit.
   DosFreeSeg (SELECTOROF (pprefList));
   WinDestroyWindow (hwndPrefFrame);
   HelpTerminate (hab, hmq);
   WinDestroyMsgQueue (hmq);
   WinTerminate (hab);
   DosExit (EXIT_PROCESS, 0);
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//   P R E F E R E N C E   C L I E N T   W I N D O W   P R O C E D U R E    //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

MRESULT CALLBACK PrefWndProc (HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2) {

   // PrefWndProc is the main client window procedure.

   static HWND hwndMenu;

   HPS hps;
   BOOL fResult;
   USHORT usResult;
   SHORT iprefSave;

   switch (msg) {

      case WM_CREATE:
         // Get the top-level menu window handle.
         hwndMenu = WinWindowFromID (WinQueryWindow (hwnd, QW_PARENT, FALSE),
                                     FID_MENU);
         return MRFROMSHORT (FALSE);

      case WM_INITMENU:

         switch (SHORT1FROMMP (mp1)) {

            case IDM_PREFMENU:
               // Disable the Restore and Delete menus if the pref list is empty.
               WinSendMsg (hwndMenu, MM_SETITEMATTR,
                           MPFROM2SHORT (IDM_RESTORE, TRUE),
                           MPFROM2SHORT (MIA_DISABLED, fEmptyPrefList ?
                                         MIA_DISABLED : FALSE));

               WinSendMsg (hwndMenu, MM_SETITEMATTR,
                           MPFROM2SHORT (IDM_DELETE, TRUE),
                           MPFROM2SHORT (MIA_DISABLED, fEmptyPrefList ?
                                         MIA_DISABLED : FALSE));

               // Disable the Undo and Save menus if no pref has been restored.
               WinSendMsg (hwndMenu, MM_SETITEMATTR,
                           MPFROM2SHORT (IDM_UNDO, TRUE),
                           MPFROM2SHORT (MIA_DISABLED, *szCurrentPref ?
                                         FALSE : MIA_DISABLED));

               WinSendMsg (hwndMenu, MM_SETITEMATTR,
                           MPFROM2SHORT (IDM_SAVE, TRUE),
                           MPFROM2SHORT (MIA_DISABLED, *szCurrentPref ?
                                         FALSE : MIA_DISABLED));

               return MRFROMSHORT (FALSE);
         }
         break;

      case WM_COMMAND:

         switch (COMMANDMSG(&msg)->cmd) {

            case IDM_RESTORE:
               // Let the user select a preference from the preference list.
               fResult = WinDlgBox (HWND_DESKTOP, hwnd, RestoreDlgProc,
                                    0, IDD_RESTORE, NULL);
               // Change system colors and values if a preference was selected.
               if (fResult)
                  ChangeCurrentPref ();
               return MRFROMSHORT (FALSE);

            case IDM_UNDO:
               // Swap the previous and current preference.
               iprefSave = iprefCurrent;
               iprefCurrent = iprefPrevious;
               iprefPrevious = iprefSave;
               ChangeCurrentPref ();
               return MRFROMSHORT (FALSE);

            case IDM_SAVE:
               // Replace the current pref list entry with current values.
               GetPrefValues (iprefCurrent);
               // Write the entire pref list.
               WritePrefFile ();
               fUnsavedChanges = FALSE;
               return MRFROMSHORT (FALSE);

            case IDM_SAVEAS:
               // Get a name for this preference from the user.
               fResult = WinDlgBox (HWND_DESKTOP, hwnd, SaveAsDlgProc,
                                    0, IDD_SAVEAS, NULL);
               // If a name was entered, save the preference.
               if (fResult) {
                  SaveNewPref (hwnd);
                  fUnsavedChanges = FALSE;
               }
               return MRFROMSHORT (FALSE);

            case IDM_DELETE:
               // Let the user select preferences to delete from the pref list.
               WinDlgBox (HWND_DESKTOP, hwnd, DeleteDlgProc, 0, IDD_DELETE, NULL);
               return MRFROMSHORT (FALSE);

            case IDM_ABOUT:
               // Display a dialog box giving information about the program.
               WinDlgBox (HWND_DESKTOP, hwnd, AboutDlgProc, 0, IDD_ABOUT, NULL);
               return MRFROMSHORT (FALSE);
         }
         break;

      case WM_PAINT:
         // Erase the client window to the background color.
         hps = WinBeginPaint (hwnd, NULL, NULL);
         GpiErase (hps);
         WinEndPaint (hps);
         return MRFROMSHORT (FALSE);

      case WM_CLOSE:
         // Ask whether to save changes if some remain unsaved.
         if (fUnsavedChanges) {
            usResult = WinMessageBox (HWND_DESKTOP, hwnd,
                                      "Save preference list changes?",
                                      "PM Preference", 0,
                                      MB_ICONQUESTION | MB_YESNOCANCEL);
            if (usResult == MBID_CANCEL)
               return MRFROMSHORT (FALSE);
            if (usResult == MBID_YES)
               WritePrefFile ();
         }
         break;
   }
   return WinDefWindowProc (hwnd, msg, mp1, mp2);
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//                     D I A L O G   P R O C E D U R E S                    //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

MRESULT CALLBACK RestoreDlgProc (HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2) {

   // RestoreDlgProc is the dialog procedure for the Restore dialog box.

   CHAR szBuffer [MAX_NAME_SIZE];
   SHORT sSelect;
   SHORT sResult;
   BOOL fTemp;

   switch (msg) {

      case WM_INITDLG:
         // Set up the preference list and set the "make permanent" check box.
         FillPrefListBox (hwnd, IDD_PREFLIST);
         // Select the first item in the list.
         WinSendDlgItemMsg (hwnd, IDD_PREFLIST, LM_SELECTITEM, 0,
                            MPFROMSHORT (TRUE));
         WinSendDlgItemMsg (hwnd, IDD_PERMANENT, BM_SETCHECK,
                            MPFROMSHORT (fPermanentPref), NULL);
         // Center the dialog box within the client window.
         CenterDialogBox (hwnd);
         return MRFROMSHORT (FALSE);

      case WM_CONTROL:
         if (SHORT1FROMMP (mp1) == IDD_PREFLIST) {
            // Perform "OK" processing if a selection was double-clicked.
            if (SHORT2FROMMP (mp1) == LN_ENTER) {
               WinSendMsg (hwnd, WM_COMMAND, MPFROM2SHORT (DID_OK, 0),
                           MPFROM2SHORT (CMDSRC_PUSHBUTTON, TRUE));
               return MRFROMSHORT (FALSE);
            }
         }
         break;

      case WM_COMMAND:

         switch (COMMANDMSG(&msg)->cmd) {

            case DID_OK:
            case IDD_ORIGINAL:
               // Get the state of the "make permanent" check box.
               fTemp = SHORT1FROMMR (WinSendDlgItemMsg (hwnd,
                                              IDD_PERMANENT, BM_QUERYCHECK,
                                              NULL, NULL));

              if (COMMANDMSG(&msg)->cmd == IDD_ORIGINAL) {
                  // Restore the original preference (index number 0).
                  iprefPrevious = iprefCurrent;
                  iprefCurrent = 0;
                  fPermanentPref = fTemp;
                  WinDismissDlg (hwnd, TRUE);
                  return MRFROMSHORT (FALSE);
               }

               // Get the current selection.
               sSelect = SHORT1FROMMR (WinSendDlgItemMsg (hwnd, IDD_PREFLIST,
                                                          LM_QUERYSELECTION,
                                                          0, 0));

               WinSendDlgItemMsg (hwnd, IDD_PREFLIST, LM_QUERYITEMTEXT,
                                  MPFROM2SHORT (sSelect, sizeof szBuffer),
                                  MPFROMP (szBuffer));

               // Find the preference index from its name.
               sResult = FindPrefName (szBuffer);
               if (sResult < 0) {
                  // We can't find the preference, so cancel the restore.
                  WinDismissDlg (hwnd, FALSE);
                  return MRFROMSHORT (FALSE);
               }
               else {
                  // Set the current preference index to the new selection.
                  iprefPrevious = iprefCurrent;
                  iprefCurrent = sResult;
                  fPermanentPref = fTemp;
                  WinDismissDlg (hwnd, TRUE);
                  return MRFROMSHORT (FALSE);
               }
               break;

            case DID_CANCEL:
               WinDismissDlg (hwnd, FALSE);
               return MRFROMSHORT (FALSE);
               break;
         }
         break;
   }
   return WinDefDlgProc (hwnd, msg, mp1, mp2);
}


MRESULT CALLBACK SaveAsDlgProc (HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2) {

   // SaveAsDlgProc is the dialog procedure for the Save As dialog box.

   switch (msg) {

      case WM_INITDLG:
         // Set the maximum entry length to the size of our buffer.
         WinSendDlgItemMsg (hwnd, IDD_EDITSAVE, EM_SETTEXTLIMIT,
                            MPFROMSHORT (sizeof szNewPref), 0);
         // Set the current name as the default entry.
         WinSetWindowText (WinWindowFromID (hwnd, IDD_EDITSAVE),
                           szCurrentPref);
         // Select the default entry.
         WinSendDlgItemMsg (hwnd, IDD_EDITSAVE, EM_SETSEL,
                            MPFROM2SHORT (0, strlen (szCurrentPref)), 0);
         // Center the dialog box within the client window.
         CenterDialogBox (hwnd);
         return MRFROMSHORT (FALSE);

      case WM_COMMAND:

         switch (COMMANDMSG(&msg)->cmd) {

            case DID_OK:
               // Get the entry text.
               WinQueryWindowText (WinWindowFromID (hwnd, IDD_EDITSAVE),
                                   sizeof szNewPref, szNewPref);
               // Cancel if no name was entered.
               if (*szNewPref == EOS)
                  WinDismissDlg (hwnd, FALSE);
               else
                  WinDismissDlg (hwnd, TRUE);
               return MRFROMSHORT (FALSE);
               break;

            case DID_CANCEL:
               WinDismissDlg (hwnd, FALSE);
               return MRFROMSHORT (FALSE);
               break;
         }
         break;
   }
   return WinDefDlgProc (hwnd, msg, mp1, mp2);
}


MRESULT CALLBACK DeleteDlgProc (HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2) {

   // DeleteDlgProc is the dialog procedure for the Delete dialog box.

   CHAR szBuffer [MAX_NAME_SIZE];
   SHORT sSelect;
   SHORT sResult;
   SHORT ipref;

   switch (msg) {

      case WM_INITDLG:
         // Set up the preference list.
         FillPrefListBox (hwnd, IDD_DELETELIST);
         // Center the dialog box within the client window.
         CenterDialogBox (hwnd);
         return MRFROMSHORT (FALSE);

      case WM_CONTROL:
         if (SHORT1FROMMP (mp1) == IDD_DELETELIST) {
            // Perform "OK" processing if a selection was double-clicked.
            if (SHORT2FROMMP (mp1) == LN_ENTER) {
               WinSendMsg (hwnd, WM_COMMAND, MPFROM2SHORT (DID_OK, 0),
                           MPFROM2SHORT (CMDSRC_PUSHBUTTON, TRUE));
               return MRFROMSHORT (FALSE);
            }
         }
         break;

      case WM_COMMAND:

         switch (COMMANDMSG(&msg)->cmd) {

            case DID_OK:
               // Delete each selection on the delete list.
               sSelect = LIT_FIRST;
               while (TRUE) {
                  // Get the list box index of the next selection.
                  sSelect = SHORT1FROMMR (WinSendDlgItemMsg (hwnd,
                                                             IDD_DELETELIST,
                                                             LM_QUERYSELECTION,
                                                             MPFROMSHORT (sSelect),
                                                             0));
                  // Stop when there are no more selections.
                  if (sSelect == LIT_NONE)
                     break;

                  // Get the name of the selected preference.
                  WinSendDlgItemMsg (hwnd, IDD_DELETELIST, LM_QUERYITEMTEXT,
                                     MPFROM2SHORT (sSelect, sizeof szBuffer),
                                     MPFROMP (szBuffer));

                  // Find the preference index from its name.
                  sResult = FindPrefName (szBuffer);
                  if (sResult > 0) {

                     // If the current preference is deleted, then return
                     // the current preference to the initial state.
                     if (sResult == iprefCurrent) {
                        iprefPrevious = 0;
                        iprefCurrent = 0;
                        *szCurrentPref = EOS;
                        WinSetWindowText (hwndPrefFrame, szPrefTitle);
                     }

                     // If the previous preference is deleted, then
                     // reset it to the original preference.
                     if (sResult == iprefPrevious)
                        iprefPrevious = 0;

                     // Delete the indicated preference from the pref list.
                     for (ipref = sResult + 1; ipref < cpref; ipref++) {

                        pprefList [ipref-1] = pprefList [ipref];

                        // Shift up the current and previous preference.
                        if (ipref == iprefCurrent)
                           iprefCurrent--;
                        else
                           if (ipref == iprefPrevious)
                              iprefPrevious--;
                     }

                     cpref--;
                     fUnsavedChanges = TRUE;
                  }
               }
               // Check if the list is completely empty now.
               if (cpref <= 1)
                  fEmptyPrefList = TRUE;

               WinDismissDlg (hwnd, TRUE);
               return MRFROMSHORT (FALSE);
               break;

            case DID_CANCEL:
               WinDismissDlg (hwnd, FALSE);
               return MRFROMSHORT (FALSE);
               break;
         }
         break;
   }
   return WinDefDlgProc (hwnd, msg, mp1, mp2);
}


MRESULT CALLBACK AboutDlgProc (HWND hwnd, USHORT msg, MPARAM mp1, MPARAM mp2) {

   // AboutDlgProc is the dialog procedure for the About PM Preference dialog box.

   switch (msg) {

      case WM_INITDLG:
         // Center the dialog box within the client window.
         CenterDialogBox (hwnd);
         return MRFROMSHORT (FALSE);

      case WM_COMMAND:

         switch (COMMANDMSG(&msg)->cmd) {

            case DID_OK:
            case DID_CANCEL:
               WinDismissDlg (hwnd, TRUE);
               return MRFROMSHORT (FALSE);
         }
      break;
   }
   return WinDefDlgProc (hwnd, msg, mp1, mp2);
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//    P M   P R E F E R E N C E   L O W - L E V E L   F U N C T I O N S     //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

VOID GetPrefValues (SHORT ipref) {

   // Query PM for system values and save them in our preference list.

   SHORT sIndex;
   USHORT iValue;
   LONG lSysConst;

   // Visit each system color and system value.
   for (sIndex = 0; sIndex < NUM_SETTINGS; sIndex++) {

      // Find the appropriate system value constant.
      lSysConst = aprefindex [sIndex].lSysConst;
      iValue = aprefindex [sIndex].iValue;

      // Store a system color or system value in our preference list.

      switch (aprefindex [sIndex].usType) {

         case PI_COLOR:
            // Only do the extra OS/2 1.2 colors if we're running 1.2 or above.
            if (fExtraValues || iValue > 14)
               pprefList [ipref].alColor [iValue] =
                  WinQuerySysColor (HWND_DESKTOP, (COLOR) lSysConst, 0L);
            break;

         case PI_VALUE:
            // Only do the extra OS/2 1.2 values if we're running 1.2 or above.
            if (fExtraValues || iValue < VALUE_BOUNDARY)
               pprefList [ipref].alValue [iValue] =
                  WinQuerySysValue (HWND_DESKTOP, (SHORT) lSysConst);
            break;
      }
   }
}


VOID SetPrefValues (SHORT ipref) {

   // Set system colors and system values for the indicated preference.

   SHORT sIndex;
   SHORT iValue;
   SHORT isSysValue;
   BOOL fStatus;
   COLOR clrFirst;
   ULONG cclr;
   SHORT iclr;

   // Do the extra OS/2 1.2 colors only if we're running 1.2 or above.

   if (fExtraValues) {
      clrFirst = SYSCLR_BUTTONLIGHT;
      cclr = NUM_COLORS;
      iclr = 0;
   }
   else {
      clrFirst = SYSCLR_WINDOWSTATICTEXT;
      cclr = NUM_COLORS - COLOR_BOUNDARY;
      iclr = COLOR_BOUNDARY;
   }

   // Set all the system colors as a group.

   fStatus = WinSetSysColors (HWND_DESKTOP, 0L, LCOLF_CONSECRGB,
                              clrFirst, cclr,
                              (PCOLOR) (pprefList[ipref].alColor + iclr));
   if (!fStatus)
      DisplayError ();

   // Set the system values.

   for (sIndex = 0; sIndex < NUM_SETTINGS; sIndex++) {

      if (aprefindex [sIndex].usType == PI_VALUE) {

         iValue = aprefindex [sIndex].iValue;

         // Only set the extra 1.2 values if we're running OS/2 1.2 or above.
         if (fExtraValues || iValue < VALUE_BOUNDARY) {

            isSysValue = (SHORT) aprefindex [sIndex].lSysConst;
            SetValue (isSysValue, pprefList [ipref].alValue [iValue]);

            // Force the x and y border width to the same value.
            if (isSysValue == SV_CXSIZEBORDER)
               SetValue (SV_CYSIZEBORDER, pprefList [ipref].alValue [iValue]);
         }
      }
   }
}


VOID ReadPrefFile (VOID) {

   // Read each preference that was saved in the pmpref.ini file.

   FILE *hfPref;
   CHAR achBuf [256];
   CHAR *pch;
   CHAR *pchTerm;
   SHORT ipref;
   USHORT iValue;
   SHORT sResult;
   SHORT sIndex;

   // The lColor structure holds a color value that will fit in a LONG.
   // We have to use this instead of a MAKETYPE from RGB to LONG so we
   // can explicitly set the bFiller byte to zero as required by
   // WinSetSysColors.

   struct {
      BYTE bBlue;
      BYTE bGreen;
      BYTE bRed;
      BYTE bFiller;
   } lColor;

   // We start at index 1 because the original settings are stored in index 0.
   cpref = ipref = 1;

   // Search the data path for the preference file.
   sResult = DosSearchPath (SEARCH_CUR_DIRECTORY | SEARCH_ENVIRONMENT,
                            "DPATH", PREF_FILE, szPrefFileName,
                            sizeof szPrefFileName);
   if (sResult != 0) {
      // The pref file was not found.  Use the default name.
      strcpy (szPrefFileName, PREF_FILE);
      return;
   }

   // Open the preference file.
   // If it doesn't exist, return with an empty pref list.
   if ((hfPref = fopen (szPrefFileName, "r")) == NULL)
      return;

   // Read each line of the preference file.
   while (fgets (achBuf, sizeof achBuf, hfPref) != NULL) {

      pch = StrTrimLeft (achBuf);
      // Skip blank lines.
      if (*pch == EOS)
         continue;

      // Start a new preference if the line indicates a preference name
      // using the form "[preference name]".

      if (*pch == '[') {

         // Get the preference name (without delimiting brackets).
         pch = StrTrimLeft (++pch);
         pchTerm = strchr (pch, ']');
         if (pchTerm != NULL)
            *pchTerm = EOS;
         StrTrimRight (pch);

         // Check if a preference by this name already exists.
         sResult = FindPrefName (pch);

         if (sResult < 0) {
            // The name was not found, so start a new preference.
            if (++cpref > sMaxPref) {
               if (!PrefAlloc ()) {
                  AbortPgm ("Not enough memory");
                  return;
               }
            }
            ipref = cpref - 1;
         }

         else {
            // Replace an existing preference.
            ipref = sResult;
         }

         // Set the new preference structure to PM defaults.
         pprefList [ipref] = prefDefault;
         strncpy (pprefList [ipref].szPrefName, pch, MAX_NAME_SIZE);
         continue;
      }

      // This is not a new preference, so fill in a preference value.
      // All key / value combinations are of the form "key: value".
      pchTerm = strchr (pch, ':');
      if (pchTerm == NULL)
         continue;
      *pchTerm = EOS;
      pch = StrTrimLeft (StrTrimRight (pch));

      // Using the key string, find the PREF structure offset for this value.
      sIndex = FindPrefKey (pch);
      // If we can't find the key, skip this value.
      if (sIndex == -1)
         continue;

      // Set the value for the indicated key.
      iValue = aprefindex [sIndex].iValue;
      if (aprefindex [sIndex].usType == PI_COLOR) {
         // Set red, green, and blue components of a PM color.
         pch += strlen (pch) + 1;
         pch = StrTrimLeft (pch);
         lColor.bRed = (BYTE) atoi (pch);
         while (*pch == '-' || isdigit (*pch))
            *pch++;
         pch = StrTrimLeft (pch);
         lColor.bGreen = (BYTE) atoi (pch);
         while (*pch == '-' || isdigit (*pch))
            *pch++;
         pch = StrTrimLeft (pch);
         lColor.bBlue = (BYTE) atoi (pch);
         lColor.bFiller = 0;
         pprefList [ipref].alColor [iValue] = MAKETYPE (lColor, LONG);
      }
      else {
         // Set a single value.
         pch += strlen (pch) + 1;
         pprefList [ipref].alValue [iValue] = atol (pch);
      }
   }
   fclose (hfPref);
}


VOID WritePrefFile (VOID) {

   // Write each each preference to the pmpref.ini file.

   FILE *hfPref;
   CHAR achBuf [256];
   SHORT ipref;
   USHORT iValue;
   SHORT sIndex;
   RGB rgb;

   // Open the preference file.
   if ((hfPref = fopen (szPrefFileName, "w")) == NULL) {
      sprintf (achBuf, "The preference file %s cannot be written.",
               szPrefFileName);
      AbortPgm (achBuf);
   }
   ShowHourglassPointer ();

   // Write each line of the preference file.
   for (ipref = 1; ipref < cpref; ipref++) {

      // Write the preference name.
      sprintf (achBuf, "[%s]\n", pprefList [ipref].szPrefName);
      fputs (achBuf, hfPref);

      // Write each preference key:value combination.
      for (sIndex = 0; sIndex < NUM_SETTINGS; sIndex++) {

         // Write the key.
         fputs (aprefindex [sIndex].szKeyName, hfPref);

         // Write the value.
         iValue = aprefindex [sIndex].iValue;

         switch (aprefindex [sIndex].usType) {
            case PI_COLOR:
               rgb = MAKETYPE (pprefList [ipref].alColor [iValue], RGB);
               sprintf (achBuf, ": %d %d %d\n", rgb.bRed, rgb.bGreen, rgb.bBlue);
               break;
            case PI_VALUE:
               sprintf (achBuf, ": %d\n", pprefList [ipref].alValue [iValue]);
               break;
         }

         fputs (achBuf, hfPref);
      }
      fputs ("\n", hfPref);
   }

   fclose (hfPref);
   ShowNormalPointer ();
}


VOID WriteIni (SHORT ipref) {

   // Write current color preferences and system values to the os2.ini file.

   SHORT sIndex;
   USHORT iValue;
   CHAR achBuf [256];
   PSZ pszKeyName;
   RGB rgb;

   for (sIndex = 0; sIndex < NUM_SETTINGS; sIndex++) {

      // Get a pointer to the key name and the index within alColor or alValue.
      pszKeyName = aprefindex [sIndex].szKeyName;
      iValue = aprefindex [sIndex].iValue;

      switch (aprefindex [sIndex].usType) {

         case PI_COLOR:
            // Convert the RGB values to a string and write it to os2.ini.
            // Only do the extra OS/2 1.2 colors if we're running 1.2 or above.
            if (fExtraValues || iValue > 14) {
               rgb = MAKETYPE (pprefList [ipref].alColor [iValue], RGB);
               sprintf (achBuf, "%d %d %d", rgb.bRed, rgb.bGreen, rgb.bBlue);
               WinWriteProfileString (hab, "PM_Colors", pszKeyName, achBuf);
            }
            break;

         case PI_VALUE:
            // Convert the system value to a string and write it to os2.ini.
            // Only do the extra OS/2 1.2 values if we're running 1.2 or above.
            if (fExtraValues || iValue < VALUE_BOUNDARY) {
               ltoa (pprefList [ipref].alValue [iValue], achBuf, 10);
               WinWriteProfileString (hab, "PM_ControlPanel", pszKeyName, achBuf);
            }
            break;
      }
   }
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//       P M   P R E F E R E N C E   U T I L I T Y   F U N C T I O N S      //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

VOID ChangeCurrentPref (VOID) {

   // Set the new current preference indicated by iprefCurrent.
   // This function combines calls to SetPrefValues and WriteIni.
   // It is called when the user specifies a preference name on
   // the command line or selects Restore or Undo from the menu.

   // Tell the system about the new system colors and values.
   ShowHourglassPointer ();
   SetPrefValues (iprefCurrent);

   // Write colors and values to the os2.ini file if the
   // "make permanent" box was checked.
   if (fPermanentPref)
      WriteIni (iprefCurrent);

   if (iprefCurrent == 0) {
      // This is the original preference.
      // Clear the preference name in the title bar.
      *szCurrentPref = EOS;
      WinSetWindowText (WinQueryWindow (hwndPref, QW_PARENT, FALSE),
                        szPrefTitle);
   }
   else {
      // Set the preference name in the title bar text.
      strcpy (szCurrentPref, pprefList [iprefCurrent].szPrefName);
      strcpy (szTitlebar, szPrefTitle);
      strcat (szTitlebar, " - ");
      strcat (szTitlebar, szCurrentPref);
      WinSetWindowText (WinQueryWindow (hwndPref, QW_PARENT, FALSE),
                        szTitlebar);
   }

   ShowNormalPointer ();
}


VOID SaveNewPref (HWND hwnd) {

   // Save the current preference settings to the preference file
   // using the name given in szNewPref.

   // This function combines calls to GetPrefValues and WritePrefFile.
   // It is called when the user selects Save As from the menu.

   SHORT sResult;
   CHAR achBuf [sizeof szNewPref + 32];

   // Is this name already used by an existing preference?
   sResult = FindPrefName (szNewPref);
   if (sResult != -1) {

      // The name is already used, so confirm the overwrite.
      sprintf (achBuf, "Replace existing '%s'?", szNewPref);
      if (WinMessageBox (HWND_DESKTOP, hwnd, achBuf, "PM Preference", 0,
                         MB_ICONQUESTION | MB_YESNO) == MBID_NO)
         return;
      iprefCurrent = sResult;
   }

   else {
      // The name is new, so add a new name to the list.
      if (++cpref > sMaxPref) {
         if (!PrefAlloc ()) {
            AbortPgm ("Not enough memory");
            return;
         }
      }
      iprefCurrent = cpref - 1;
   }


   // Replace the current pref list entry with current values.
   GetPrefValues (iprefCurrent);

   // Save the new preference name.
   strcpy (pprefList [iprefCurrent].szPrefName, szNewPref);

   // Set the preference name in the title bar text.
   strcpy (szCurrentPref, pprefList [iprefCurrent].szPrefName);
   strcpy (szTitlebar, szPrefTitle);
   strcat (szTitlebar, " - ");
   strcat (szTitlebar, szCurrentPref);
   WinSetWindowText (WinQueryWindow (hwndPref, QW_PARENT, FALSE),
                     szTitlebar);

   // Write the entire pref list.
   WritePrefFile ();
}


VOID SetValue (SHORT isSysValue, LONG lValue) {

   BOOL fStatus;

   // Set the indicated system value.
   fStatus = WinSetSysValue (HWND_DESKTOP, isSysValue, lValue);
   if (!fStatus)
      DisplayError ();
   else
      // Notify all frame windows that this value has changed.
      WinBroadcastMsg (HWND_DESKTOP, WM_SYSVALUECHANGED,
                       MPFROMSHORT (isSysValue), MPFROMSHORT (isSysValue),
                       BMSG_FRAMEONLY | BMSG_SEND);

}


VOID FillPrefListBox (HWND hwnd, USHORT idListBox) {

   // Fill the preference list box starting with the first preference
   // that was read from the preference file (index 1).

   SHORT ipref = 1;

   WinSendDlgItemMsg (hwnd, idListBox, LM_DELETEALL, NULL, NULL);

   while (ipref < cpref) {
      WinSendDlgItemMsg (hwnd, idListBox, LM_INSERTITEM,
                         MPFROM2SHORT (LIT_SORTASCENDING, 0),
                         MPFROMP (pprefList [ipref].szPrefName));
      ipref++;
   }
}


SHORT FindPrefKey (CHAR *pchKey) {

   // Return the index in aprefindex of the preference setting
   // with name pchKey.  Return -1 if the name is not found.

   SHORT sIndex = 0;

   // Compare the key string to each key string we know about.
   while (sIndex < NUM_SETTINGS) {
      if (strcmp (pchKey, aprefindex [sIndex].szKeyName) == 0)
         break;
      sIndex++;
   }

   // Indicate failure if we didn't find a match.
   if (sIndex >= NUM_SETTINGS)
      sIndex = -1;

   return (sIndex);
}


SHORT FindPrefName (CHAR *pchPrefName) {

   // Compare the preference name to each name in the preference list
   // starting with the first preference read from the file (index 1).
   // Note that names are case-insensitive, even though we allow mixed-
   // case names to be entered and store them that way.  Return -1 if
   // the name is not found.

   SHORT ipref = 1;

   while (ipref < cpref) {
      if (stricmp (pchPrefName, pprefList [ipref].szPrefName) == 0)
         break;
      ipref++;
   }

   // Indicate failure if we didn't find a match.
   if (ipref >= cpref)
      ipref = -1;

   return (ipref);
}


BOOL PrefAlloc (VOID) {

   // Allocate memory to store preferences.  The memory allocated
   // is increased by an amount equal to PREF_ALLOC preferences
   // each time this function is called.  PrefAlloc returns TRUE
   // if the memory was allocated, FALSE if it wasn't.  The variable
   // sMaxPref always indicates the number of preferences for which
   // memory has been allocated.

   static SEL selPrefList;

   // If ppprefList is still NULL, this is the first allocation.
   if (pprefList == NULL) {
      if (DosAllocSeg (PREF_ALLOC * sizeof (PREF), &selPrefList, 0))
         return FALSE;
      pprefList = MAKEP (selPrefList, 0);
   }

   else {
      // Add room for PREF_ALLOC more preferences.
      if (DosReallocSeg ((sMaxPref + PREF_ALLOC) * sizeof (PREF),
                         selPrefList))
         return FALSE;
   }

   sMaxPref += PREF_ALLOC;
   return TRUE;
}


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//             G E N E R A L   U T I L I T Y   F U N C T I O N S            //
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

VOID AbortPgm (CHAR *szErrorMessage) {

   // Display an error message, wait for confirmation, then stop the program.

   WinMessageBox (HWND_DESKTOP, hwndPref, szErrorMessage, "PM Preference Error",
                  0, MB_OK | MB_ICONEXCLAMATION | MB_MOVEABLE);

   WinSendMsg (hwndPref, WM_CLOSE, 0, 0);
}


VOID DisplayMsg (CHAR *szErrorMessage) {

   // Display an informational message and wait for confirmation.

   WinMessageBox (HWND_DESKTOP, hwndPref, szErrorMessage, "PM Preference",
                  0, MB_OK | MB_NOICON | MB_MOVEABLE);
}


VOID DisplayError (VOID) {

   // Retrieve the last PM error condition then display the error code.

   ERRORID idError = WinGetLastError (hab);
   char achBuf [64];

   sprintf (achBuf, "Error number %d has occurred.", ERRORIDERROR (idError));
   WinMessageBox (HWND_DESKTOP, hwndPref, achBuf, "PM Preference Error",
                  0, MB_OK | MB_ICONEXCLAMATION | MB_MOVEABLE);
}


CHAR *StrTrimLeft (CHAR *pch) {

   // Return a pointer to the first non-space character in the pch string.

   while (isspace (*pch))
      pch++;
   return (pch);
}


CHAR *StrTrimRight (CHAR *pch) {

   // Null-terminate the pch string after the last non-space character.

   CHAR *pchTerm = pch + strlen (pch) - 1;
   while (isspace (*pchTerm))
      pchTerm--;
   *++pchTerm = EOS;
   return (pch);
}


VOID ShowHourglassPointer (VOID) {

   // Display the universally-reviled hourglass pointer.

   WinSetPointer (HWND_DESKTOP, WinQuerySysPointer (HWND_DESKTOP, SPTR_WAIT,
                                                    FALSE));
}


VOID ShowNormalPointer (VOID) {

   // Display the normal arrow pointer.

   WinSetPointer (HWND_DESKTOP, WinQuerySysPointer (HWND_DESKTOP, SPTR_ARROW,
                                                    FALSE));
}


VOID CenterDialogBox (HWND hwnd) {

   // Position a dialog box at the center of the client window.

   // Since we must set the position of the dialog window relative to its
   // parent, the desktop window, we must first find the position of the
   // client window relative to the desktop window.  We add the coordinates
   // of the frame window (relative to the desktop) to the coordinates of
   // the client window (relative to the frame window).

   HWND hwndFrame, hwndClient;
   SWP swpFrame, swpClient;
   RECTL rclDialog;
   SHORT sWidth, sHeight;
   SHORT sBLCx, sBLCy;

   SHORT cxDeskTop = (SHORT) WinQuerySysValue (HWND_DESKTOP, SV_CXSCREEN);
   SHORT cyDeskTop = (SHORT) WinQuerySysValue (HWND_DESKTOP, SV_CYSCREEN);

   // Get the frame window and client window handles.
   hwndFrame = WinQueryWindow (hwnd, QW_OWNER, FALSE);
   hwndClient = WinWindowFromID (hwndFrame, FID_CLIENT);

   // Get the boundary and position of the frame and client windows.
   WinQueryWindowPos (hwndFrame, &swpFrame);
   WinQueryWindowPos (hwndClient, &swpClient);

   // Get the boundary rectangle of the dialog window.
   WinQueryWindowRect (hwnd, &rclDialog);

   // Get the height and width of the dialog box.
   sHeight = (SHORT) (rclDialog.yTop - rclDialog.yBottom);
   sWidth = (SHORT) (rclDialog.xRight - rclDialog.xLeft);

   // Set the coordinates for the bottom left corner of the dialog.
   sBLCx = swpFrame.x + swpClient.x + (swpClient.cx - sWidth) / 2;
   sBLCy = swpFrame.y + swpClient.y + (swpClient.cy - sHeight) / 2;

   // Ensure that the dialog window is entirely on the screen.
   sBLCx = max (0, min (sBLCx, cxDeskTop - sWidth));
   sBLCy = max (0, min (sBLCy, cyDeskTop - sHeight));

   // Move the dialog box window to its new position.
   WinSetWindowPos (hwnd, HWND_TOP, sBLCx, sBLCy, 0, 0, SWP_MOVE);
}
